package ch.epfl.ppo.collections;

import java.util.Iterator;
import java.util.NoSuchElementException;

public class SimpleHashSet4<E> implements Set<E> {
    private static final double IDEAL_LOAD_FACTOR = 0.75;
    private static final double MAX_LOAD_FACTOR = 1;
    private static final int MIN_CAPACITY = 10;

    private int size;
    private ListSet<E>[] table;

    public SimpleHashSet4() {
        this.size = 0;
        this.table = newTable(MIN_CAPACITY);
    }

    @Override
    public boolean isEmpty() {
        return size == 0;
    }

    @Override
    public int size() {
        return size;
    }

    @Override
    public void add(E newElem) {
        ListSet<E> list = listFor(table, newElem);
        size -= list.size();
        list.add(newElem);
        size += list.size();

        double loadFactor = (double)size / (double)table.length;
        if (loadFactor > MAX_LOAD_FACTOR)
            rehash();
    }

    private void rehash() {
        ListSet<E>[] newTable = newTable((int)(size / IDEAL_LOAD_FACTOR));
        for (E elem: this)
            listFor(newTable, elem).add(elem);
        table = newTable;
    }

    @Override
    public void remove(E elem) {
        ListSet<E> list = listFor(table, elem);
        size -= list.size();
        list.remove(elem);
        size += list.size();
    }

    @Override
    public boolean contains(E elem) {
        return listFor(table, elem).contains(elem);
    }

    @Override
    public Iterator<E> iterator() {
        return new Iterator<E>() {
            private int remaining = size;
            private int i = 0;
            private Iterator<E> it = table[i].iterator();

            @Override
            public boolean hasNext() {
                return remaining > 0;
            }

            @Override
            public E next() {
                if (! hasNext())
                    throw new NoSuchElementException();
                --remaining;
                while (! it.hasNext())
                    it = table[++i].iterator();
                return it.next();
            }

            @Override
            public void remove() {
                it.remove();
                --size;
            }
        };
    }

    private static <E> ListSet<E>[] newTable(int capacity) {
        @SuppressWarnings("unchecked")
        ListSet<E>[] table = new ListSet[capacity];
        for (int i = 0; i < table.length; ++i)
            table[i] = new ListSet<>();
        return table;
    }

    private static <E> ListSet<E> listFor(ListSet<E>[] table, E elem) {
        return table[Math.abs(elem.hashCode() % table.length)];
    }
}
